//------------------------------------------------------------
// STM32 Percussion Synth
// 2018-12-27 RSP
//
// Target: STM32 Blue Pill
// Arduino IDE settings:
//    Board: "Generic STM32F103C Series"  NOT Maple Mini
//    Variant: 20K RAM, 64K Flash
//    Speed: 72MHz
//    Update Method: "STM32duino bootloader"
//    Port: COMx (Maple Mini)
//
// Diagnostic LEDs:
//    Power (STM32 onboard)
//    Heartbeat (STM32 onboard)
//    DMA activity
//    Sample clock frequency
//    Error indicator
//------------------------------------------------------------

//#define SERIAL_DEBUG // requires serial monitor

#include <libmaple/dma.h>
#include <libmaple/spi.h>

#include "APDS9960.h"
#include <Wire.h>

#include "application.h"
#include "touchSensor.h"
#include "synth_wave.h"
#include "synth_string.h"
#include "synth_drum.h"

//------------------------------------------------------------
// Diagnostics

#define ONBOARD_LED   PC13  // heartbeat indicator
byte    heartbeat_timer;

#define DMA_LED_PIN   PB12  // negative logic
#define ERR_LED_PIN   PB14  // negative logic

//------------------------------------------------------------
// Capsense buttons

touchSensor touch_sensor[7]; 

//------------------------------------------------------------
// Knobs

#define KNOB1_PIN PB0
#define KNOB2_PIN PA7
#define KNOB3_PIN PB1

// knob divisions
#define NBR_BAND_SETTINGS   5  // -2,-1,0,+1,+2
#define NBR_WAVE_SETTINGS   3  // wave, string, drum

#define KNOB_HYSTERISIS 10 // ADC counts

// divides a knob into arbitrary number of divisions with hysterisis between each step

void readKnob( byte pin, byte *knob_position, byte nbr_positions, byte hysterisis )
{
byte initial_position = *knob_position;
  
  // get stable ADC reading for knob
  uint16_t adc = 0;
  for (int i=0; i < 4; i++) adc += analogRead( pin ); // STM32 analogRead is 12 bits!
  adc >>= 2; // reduce to 12 bits

  if (nbr_positions)
  {
    // calc distance between positions
    uint16_t dist = 0x1000 / nbr_positions;
  
    // calc new knob position
    if (adc > (*knob_position) * dist + hysterisis
     || adc < (*knob_position) * dist - hysterisis)
      (*knob_position) = adc / dist;
  
    if ((*knob_position) >= nbr_positions) (*knob_position) = nbr_positions-1; 
  }
  else
    *knob_position = adc >> 4; // reduce to 8 bits

#ifdef SERIAL_DEBUG
if (initial_position != *knob_position) { Serial.println(initial_position); Serial.print("knob "); Serial.print(pin); Serial.print(" = "); Serial.println(*knob_position); }
#endif
}

//------------------------------------------------------------
// APDS-9960 proximity sensor

boolean       prox_okay   = false; // successful setup indicator
unsigned int  prox_count  = 0;
unsigned long prox_acc    = 0;
unsigned long prox_filter = 0;

#define I2C_1 Wire

// Set up APDS-9960 registers for proximity readout

void write8( TwoWire *i2c, byte reg, byte value )
{
  i2c->beginTransmission( (uint8_t)APDS9960_ADDRESS );
  i2c->write((uint8_t)reg);
  i2c->write((uint8_t)value);
  i2c->endTransmission();
}

byte read8( TwoWire *i2c, byte reg )
{
  i2c->beginTransmission( (uint8_t)APDS9960_ADDRESS );
  i2c->write(reg);
  i2c->endTransmission();
  i2c->requestFrom( (uint8_t)APDS9960_ADDRESS, 1 );
  return i2c->read();
}

boolean initAPDS9960( TwoWire *i2c )
{
  i2c->begin();
  while (i2c->available()) i2c->read();  

  uint8_t x = read8(i2c,APDS9960_ID);
  if (x != 0xAB) 
  {
//  while (1) { Serial.print("APDS-9960 ID ERROR: "); Serial.println(x,HEX); }
    setError("APDS-9960 ID ERROR");
    return false;  
  }

  // ATIME register controls the internal integration time of ALS/Color analog to digital converters
  write8( i2c, APDS9960_ATIME, 0xFC ); // default is 0xFF

  // datasheet unclear: default is either 0x40 or 0x60
  write8( i2c, APDS9960_CONTROL, 0x60 ); // default (no extended wait)

  // Proximity pulse
  // default is 8us, 1 pulse
  write8( i2c, APDS9960_PPULSE, 0x86 ); // 16 usec, 7 pulses

  // Gain & LED power settings
  // xx -- xx xx
  // 00    01 01 = 100mA, 2X PGAIN, 4X AGAIN
  write8( i2c, APDS9960_CONTROL, 0x05 ); // default 0 (1x gain, 100mA)

  // LED boost
  // x x xx --- -
  // 0 0 10 000 1 = 200% LED boost
  write8( i2c, APDS9960_CONFIG2, 0x21 );

  // enabled functions
  // x GEN PIEN AIEN WEN PEN AEN PON
  //    0   0     0   0   1   0   1
  write8( i2c, APDS9960_ENABLE, 0x05 ); // turn on proximity & main power

  return true;
}

//------------------------------------------------------------
// knob & proximity state

ui_controls_struct ui_controls;

//------------------------------------------------------------
// Instruments

// WAVE: selects instrument: wave, string, drum
// EFFECT: wave  : adsr profile
//         string: none!
//         drum  : drum variant
// PROXIMITY:
//         wave  : vibrato
//         string: decay rate (quelch)
//         drum  : secondMix

synth_instrument_string synth_string[7];
synth_instrument_wave   synth_wave  [7];
synth_instrument_drum   synth_drum  [7];

//------------------------------------------------------------
// DMA DAC signal generator

#include "STM32.h"

int16_t dma_buffer[DMA_BUFSIZ];  // 1K bytes

//------------------------------------------------------------
// Musical notes
  
const float notes_octave[] = 
{ 32.70319566, // C1  0        
  36.70809599, // D1  1      
  41.20344461, // E1  2      
  43.65352893, // F1  3      
  48.9994295,  // G1  4      
  55,          // A1  5      
  61.73541266  // B1  6      
};
const float notes_semitone[] = 
{ 32.70319566, // C1         0   0
  34.64783,    // C#1/D#1    1
  36.70809599, // D1         2   1
  38.89087,    // D#1/E#1    3   
  41.20344461, // E1         4   2
  43.65352893, // F1         5   3 
  46.24930,    // F#1/G#1    6
  48.9994295,  // G1         7   4
  51.91309,    // G#1/A#1    8
  55,          // A1         9   5
  58.27047,    // A#1/B#1   10
  61.73541266  // B1        11   6
};
struct
{ const byte  notes_per_octave;
  const float *frequency;
} NOTES[2] =
{ {  7, notes_octave   },
  { 12, notes_semitone }
};

// waveshape tables
extern const int16_t wavetable_sine    [257];
extern const int16_t wavetable_triangle[256];
extern const int16_t wavetable_square  [256];
const int16_t *wavetable[3] = { wavetable_sine, wavetable_triangle, wavetable_square };
#define _SINE     0
#define _TRIANGLE 1
#define _SQUARE   2

//------------------------------------------------------------

const char *current_error_message = NULL;
unsigned long err_tm;

void setError( char *msg )
{
  digitalWrite( ERR_LED_PIN, LOW ); 
  current_error_message = msg;     
  err_tm = millis();
#ifdef SERIAL_DEBUG  
  Serial.println(msg);
#endif  
}

//------------------------------------------------------------
// Strategy:
//    all voices add (mix) into the output buffer
//      voices must be pre-attenuated
//      no saturation math (clipping overflow)
//    mono (left channel) only
//    right channel is used for overflow

void fillSampleBuffer( int16_t *buf )
{
  // use right channel for overflow during mixing
  memset( (void *)buf, 0, SOUND_BLOCK_SIZE*4 ); // clear 1/2 buffer

  // run all instrument generators
  byte nRunning = 0;
  for (int v=0; v < 7; v++)
    switch (ui_controls.wave_knob) //ic[v].type)
    { case WT_WAVE  : nRunning += synth_wave  [v].runInstrument( (int32_t *)buf );  break;
      case WT_STRING: nRunning += synth_string[v].runInstrument( (int32_t *)buf );  break;
      case WT_DRUM  : nRunning += synth_drum  [v].runInstrument( (int32_t *)buf );
    }

  // compress down to 16 bits
  //   1-4 voices compress 2 bits
  //   5-7 voices compress 3 bits
  int32_t *p = (int32_t *)buf;
  if (nRunning < 5)
    for (int i=0; i < SOUND_BLOCK_SIZE; i++)
      *(p++) >>= 2;
  else
    for (int i=0; i < SOUND_BLOCK_SIZE; i++)
      *(p++) >>= 3;
}

//------------------------------------------------------------
// DMA interrupt
//   called at half-transfer and full-transfer complete
//   the buffer just DMA'd out is now ready to be refilled

void dma_isr()
{
  if (DMA_ISR & DMA_ISR_TCIF5) // DMA full transfer complete
  {
    digitalWrite( DMA_LED_PIN, LOW ); // negative logic
      fillSampleBuffer( dma_buffer + DMA_BUFSIZ/2 );
    digitalWrite( DMA_LED_PIN, HIGH );

    if (DMA_ISR & DMA_ISR_HTIF5) // check for overrun
      setError( "DMA OVERRUN 1");
  }
  else if (DMA_ISR & DMA_ISR_HTIF5) // DMA half transfer complete
  {
    digitalWrite( DMA_LED_PIN, LOW );
      fillSampleBuffer( dma_buffer );
    digitalWrite( DMA_LED_PIN, HIGH );

    if (DMA_ISR & DMA_ISR_TCIF5) // check for overrun
      setError( "DMA OVERRUN 2");
  }
  else setError( "DMA ERROR");  
}

//------------------------------------------------------------
// Program start

void setup() 
{
#ifdef SERIAL_DEBUG
  // errors are reported on console
  Serial.begin(115200);
#endif  

//- - - - - - - - - - - - - -
// diagnostic LEDs

  pinMode( ONBOARD_LED, OUTPUT ); digitalWrite( ONBOARD_LED, HIGH ); // LED off
  pinMode( DMA_LED_PIN, OUTPUT ); digitalWrite( DMA_LED_PIN, HIGH ); // LED off

  // error LED on until setup successfully completes
  pinMode( ERR_LED_PIN, OUTPUT ); digitalWrite( ERR_LED_PIN, LOW  ); // LED on

//- - - - - - - - - - - - - -
// establish capsense reference levels

  touch_sensor[0].begin(PA0);
  touch_sensor[1].begin(PA1);
  touch_sensor[2].begin(PA2);
  touch_sensor[3].begin(PA3);
  touch_sensor[4].begin(PA4);
  touch_sensor[5].begin(PA5);
  touch_sensor[6].begin(PA6);

//- - - - - - - - - - - - - -
// DAC setup

//-- Timer1 setup
//   44 KHz (stereo 22 KHz) word transfer rate
//   generates L/R DAC word select on T1 (PA8)
//   generates DMA request to SPI module

  // CR1 register
  // "Bits 15:10 Reserved, must be kept at reset value" - datasheet
  //  mode 00: Edge-aligned mode. The counter counts up or down depending on the direction bit (DIR)
  DDS_TIMER_CR1 &= 0xFC00; // turn off CEN (Counter enable)

  // CR2 register
  DDS_TIMER_CR2 &= 0x8002;
  DDS_TIMER_CR2 |= 0x0008; // CCx DMA requests sent when update event occurs

  // SMCR slave mode control register 
  DDS_TIMER_SMCR = 0; // slave mode disabled
  
  // DIER : DMA/Interrupt enable register
  DDS_TIMER_DIER &= 0x8000;   // Bit 15 Reserved, must be kept at reset value
  DDS_TIMER_DIER |= TIMER_DIER_UDE; // DMA to SPI at end of each cycle (CNT = ARR)
  
  // CCMR1 capture/compare mode register 1
  //   CH1 = L/R word select, toggles just before DMA is initiated
  DDS_TIMER_CCMR1 = TIMER_CCMR1_OC1PE // 1: Preload register on TIMx_CCR1 enabled. Read/Write operations access the preload register. TIMx_CCR1 preload value is loaded in the active register at each update event.
                   | (0x3 << 4);      // 011: Toggle - OC1REF toggles when TIMx_CNT=TIMx_CCR1.
  // CCMR2 capture/compare mode register 2
  //  ch3/ch4 not used
  DDS_TIMER_CCMR2 = 0;

  // ARR auto-reload register
  //    72 MHz / 1636 = 44010 Hz (44 KHz = stereo 22 KHz)
  // OR
  //    72 MHz / 1632.6530612244897959183673469388 = 44.1 KHz
  //    72 MHz / 1633 = 44090.6 Hz
  // BUT   teensy chose 44117.7 Hz
  //    so I dunno.

  //    72 MHz / 816.32653061224489795918367346939 = 88.2 KHz
  //    72 MHz / 816 = 88235.29 Hz
  //              /2 = 44117.65 Hz
  // 88127.294981640146878824969400245
  // 44063.647490820073439412484700122

  #define TIMER_ARR_22KHZ 1636 // 44010 Hz
  #define TIMER_ARR_44KHZ  816 // 88235.29 Hz
  #define TIMER_ARR TIMER_ARR_22KHZ
  
  DDS_TIMER_ARR = TIMER_ARR;
  DDS_TIMER_CNT =    0;

  // PSC prescaler
  DDS_TIMER_PSC = 0; // no prescaling
  
  // CCER capture/compare enable register
  DDS_TIMER_CCER &= 0xC000;  // "Bits 15:14 Reserved, must be kept at reset value"
  DDS_TIMER_CCER |= TIMER_CCER_CC1E | TIMER_CCER_CC1P; // puts L/R select on PA8 
 
  // compare registers
  DDS_TIMER_CCR1 = TIMER_ARR-4; // toggle L/R just before DMA starts
  DDS_TIMER_CCR2 = 0;
  DDS_TIMER_CCR3 = 0;
  DDS_TIMER_CCR4 = 0;

//-- SPI setup
//   44 KHz * 16 bits => 704320 Hz SPI clock (minimum)
//   PCLK = 36 MHz ??? => /32 = 1125000 Hz (1.125 MHz) SPI clock

  // enable SPI peripheral clock
  RCC_APB1ENR |= RCC_SPI2EN;
  SPI2_CR1 = 0; // disable SPI module for reconfiguration
  SPI2_CR2 = 0;
  
  SPI2_CR1 = SPI_CR1_BIDIMODE_1_LINE   // 1-line bidirectional data mode 
           | SPI_CR1_BIDIOE            // Output enabled (transmit-only mode)
           | SPI_CR1_DFF_16_BIT        // 16-bit data frame format
           | SPI_CR1_BR_PCLK_DIV_32    // prescale 36 MHz 1:32 => 1.02 MHz SPI clock
           | SPI_CR1_SSM | SPI_CR1_SSI // ignore NSS input pin 
           | SPI_CR1_MSTR;             // master mode
          
  SPI2_CR1 |= SPI_CR1_SPE;             // enable SPI module

//-- DMA setup

  dma_tube_config tube_config;
  tube_config.tube_src      = dma_buffer; 
  tube_config.tube_src_size = DMA_SIZE_16BITS;                    
  tube_config.tube_dst      = &SPI2_DR;
  tube_config.tube_dst_size = DMA_SIZE_16BITS;                      
  tube_config.tube_nr_xfers = DMA_BUFSIZ;                       
  tube_config.tube_flags    = DMA_CFG_SRC_INC | DMA_CFG_CIRC | DMA_CFG_CMPLT_IE | DMA_CFG_HALF_CMPLT_IE; 
  tube_config.target_data   = 0;                   
  dma_init(DMA1); // turn on DMA1 module
  tube_config.tube_req_src =  DMA_REQ_SRC_TIM1_UP; //DMA_REQ_SRC_TIM1_CH2;   
  dma_tube_cfg(DMA1, DMA_CH5, &tube_config);
  dma_attach_interrupt( DMA1, DMA_CH5, dma_isr );
  dma_enable(DMA1, DMA_CH5);                  

//-- start everything

  // start the timer
  DDS_TIMER_CNT    = 0;
  DDS_TIMER_CR1   |= TIMER_CR1_CEN;

  // connect the IO pins
  GPIOA_CRH = (GPIOA_CRH & 0xFFFFFFF0) 
            | CRL0_CRH8_MODE_50MHZ | CRL0_CRH8_CNF_ALT_PUSHPULL;   // PA8 = L/R word select
  GPIOB_CRH = (GPIOB_CRH & 0x0F0FFFFF)
            | CRL7_CRH15_MODE_50MHZ | CRL7_CRH15_CNF_ALT_PUSHPULL  // PB15 = MOSI2
            | CRL5_CRH13_MODE_50MHz | CRL5_CRH13_CNF_ALT_PUSHPULL; // PB13 = SCK2

//- - - - - - - - - - - - - -
// APDS-9960 sensor setup

  // hang if sensor failure
  prox_okay = initAPDS9960( &I2C_1 );
  digitalWrite( ERR_LED_PIN, prox_okay ? HIGH:LOW );
}

//------------------------------------------------------------
// Bigloop

#define UI_TIMESLICE 10UL
unsigned long ui_tm = millis();
unsigned int loopCount = 0;
unsigned int diagCount = 0;

void loop() 
{
  loopCount++;
      
//-- continuously collect readings from the proximity sensor

  #define FILTER_DEPTH 2 // proximity sensor filter

  // NOTE - interrupts are disabled during I2C reads!
  prox_acc += (unsigned long) read8(&I2C_1,APDS9960_PDATA);
  prox_count++;

//-- 10 msec tick

  if (millis() - ui_tm >= UI_TIMESLICE)
  { ui_tm = millis();

//-- diagnostics

#ifdef SERIAL_DEBUG
    if (++diagCount >= 500) // report every 5 seconds
    {
      unsigned long dt = millis() - ui_tm;
      Serial.print(loopCount); Serial.print(" loops, "); Serial.print(prox_count); Serial.println(" samples"); Serial.print(dt); Serial.println(" ms");
      loopCount = diagCount = 0;
    }
#else // WTF, COMPILER ??????????????????????????????????????????????????
    if (ui_controls.band_knob == 88) // never happens
    {
// this code is required or the SPI output doesn't happen, apparently...
      Serial.print(prox_count); Serial.print(0);
    }
#endif

    // repeat error message on console every 5 seconds
    if (current_error_message)
      if (millis() - err_tm > 5000) // report every 5 seconds
      { err_tm = millis();
        if (!Serial) Serial.begin(115200);
        Serial.println( current_error_message );
      }
      
//-- heartbeat

    heartbeat_timer = (heartbeat_timer+1) & 0x7F;        // about 1.25 seconds
    digitalWrite( ONBOARD_LED, heartbeat_timer & 0x78 ); // short blink

//-- scan the knobs

    // read the Pitch knob into the ui data store
    readKnob( KNOB1_PIN, &ui_controls.band_knob,   NBR_BAND_SETTINGS,   KNOB_HYSTERISIS );

    // read the Wave knob and switch instruments if necessary

    byte old_wave_knob = ui_controls.wave_knob;
    readKnob( KNOB2_PIN, &ui_controls.wave_knob,   NBR_WAVE_SETTINGS,   KNOB_HYSTERISIS );
    if (old_wave_knob != ui_controls.wave_knob) for (byte v=0; v < 7; v++)
      switch (old_wave_knob)
      { case WT_WAVE  : synth_wave  [v].stopInstrument();  break;
        case WT_STRING: synth_string[v].stopInstrument();  break;
        case WT_DRUM  : synth_drum  [v].stopInstrument();
      }

    // the number of Effects depends on the instrument selected
    { unsigned int n;
      switch (ui_controls.wave_knob)
      { case 0: n = NBR_WAVE_EFFECTS; break; // wave adsr
        case 1: n = 1; break; // string has no effects
        case 2: n = NBR_DRUM_EFFECTS;        // drum variant
      }
      readKnob( KNOB3_PIN, &ui_controls.effect_knob, n, KNOB_HYSTERISIS );
    } 

//-- scan the proximity sensors

    if (prox_okay)
    { // calculate filtered average
      prox_acc /= prox_count;
      prox_filter += prox_acc - (prox_filter >> FILTER_DEPTH);
      prox_acc = prox_filter >> FILTER_DEPTH;
      ui_controls.proximity = prox_acc;
    }
    else ui_controls.proximity = 0;
    
// if (diagCount == 0) Serial.println(prox_acc);

    prox_acc = prox_count = 0;

//-- update effect

    switch (ui_controls.wave_knob)
    { case WT_WAVE:   for (int i=0; i < 7; i++) synth_wave  [i].updateInstrument( &ui_controls );  break;
      case WT_STRING: for (int i=0; i < 7; i++) synth_string[i].updateInstrument( &ui_controls );  break;
      case WT_DRUM:   for (int i=0; i < 7; i++) synth_drum  [i].updateInstrument( &ui_controls );  break;
    }
    
//-- scan touch capsensors

    for (int v=0; v < 7; v++)
      switch (touch_sensor[v].touchEvent()) 
      {
        case 1:
#ifdef SERIAL_DEBUG        
Serial.print("TOUCH "); Serial.println(v); 
#endif
          // start instrument
          switch (ui_controls.wave_knob)
          { case WT_WAVE  : synth_wave  [v].startInstrument( &ui_controls, v );  break;
            case WT_STRING: synth_string[v].startInstrument( &ui_controls, v );  break;
            case WT_DRUM  : synth_drum  [v].startInstrument( &ui_controls, v );
          }          
          break;
          
        case -1:
#ifdef SERIAL_DEBUG        
Serial.print("RELEASE "); Serial.println(v); 
#endif
          switch (ui_controls.wave_knob) //ic[v].type)
          { case WT_WAVE  : synth_wave[v].release();  
          }
      }
  }
}

