#include <Arduino.h>
#include "application.h"
#include "synth_drum.h"

#define K_A1 55.0
#define K_B1 61.74

#define K_A3 220.0
#define K_B3 246.94
#define K_C4 261.63
#define K_D4 293.66
#define K_E4 329.63
#define K_F4 349.23
#define K_G4 392.0
#define K_A4 440.0
#define K_B4 493.88
#define K_C5 523.25
#define K_D5 587.33
#define K_E5 659.25
#define K_F5 698.45
#define K_G5 783.99
#define K_A5 880.0
#define K_B5 987.77
#define K_C6 1046.5
#define K_D6 1174.66
#define K_E6 1318.51
#define K_F6 1396.91
#define K_G6 1567.98


const struct
{ float   frequency; // Hz
  int32_t length;    // msec
  float   secondMix; // Emulates a two-headed tom, by adding a second sine wave that is harmonized a perfect fifth above the base frequency
  float   pitchMod;  // Set the depth of envelope of the pitch, by a maximum of two octaves. Default is 0.5, with no modulation 
                     //   values above 0.5 cause the pitch to sweep downwards, values lower than 0.5 cause the pitch to sweep upwards.
} drums[NBR_DRUM_EFFECTS][7] = // 5 effects X 7 keys
{

// pew pew, proximity vary secondMix 0-1
#define PEW_MS 300
#define PEW_PMOD 1.0
  { {K_A3,PEW_MS,0.0,PEW_PMOD}, 
    {K_B3,PEW_MS,0.2,PEW_PMOD}, 
    {K_C4,PEW_MS,0.4,PEW_PMOD}, 
    {K_D4,PEW_MS,0.5,PEW_PMOD}, 
    {K_E4,PEW_MS,0.6,PEW_PMOD},
    {K_F4,PEW_MS,0.8,PEW_PMOD},
    {K_G4,PEW_MS,1.0,PEW_PMOD} }, 

// bell, proximity vary secondMix 0-1
#define BELL_MS 400
#define BELL_PMOD 0.5
  { {K_C5,BELL_MS,0.0,BELL_PMOD},
    {K_D5,BELL_MS,0.2,BELL_PMOD}, 
    {K_E5,BELL_MS,0.4,BELL_PMOD}, 
    {K_F5,BELL_MS,0.5,BELL_PMOD}, 
    {K_G5,BELL_MS,0.6,BELL_PMOD}, 
    {K_A5,BELL_MS,0.8,BELL_PMOD}, 
    {K_B5,BELL_MS,1.0,BELL_PMOD} },

// drip, proximity vary secondMix 0-1
#define DRIP_MS 150
#define DRIP_PMOD 0.0
  { {K_C5,DRIP_MS,0.0,DRIP_PMOD},
    {K_D5,DRIP_MS,0.2,DRIP_PMOD}, 
    {K_E5,DRIP_MS,0.4,DRIP_PMOD}, 
    {K_F5,DRIP_MS,0.5,DRIP_PMOD}, 
    {K_G5,DRIP_MS,0.6,DRIP_PMOD}, 
    {K_A5,DRIP_MS,0.8,DRIP_PMOD}, 
    {K_B5,DRIP_MS,1.0,DRIP_PMOD} },

// drop, proximity vary secondMix 0-1
#define DROP_MS 300
#define DROP_PMOD 0.25
  { {K_A3,DROP_MS,0.0,DROP_PMOD},
    {K_D4,DROP_MS,0.2,DROP_PMOD}, 
    {K_G4,DROP_MS,0.4,DROP_PMOD}, 
    {K_C5,DROP_MS,0.5,DROP_PMOD}, 
    {K_F5,DROP_MS,0.6,DROP_PMOD}, 
    {K_B5,DROP_MS,0.8,DROP_PMOD}, 
    {K_B5,DROP_MS,1.0,DROP_PMOD} },

// space gun effects, vary secondMix 0-1
#define SPGUN_MS 1500
#define SPGUN_F  K_A5
  {
    {K_A5,1500,0.8,0.8}, 
    {K_D6, 150,0.0,0.0}, 
    {K_C4,1500,0.0,0.55},
    {K_C6, 200,0.4,0.25},
    {K_C4, 500,0.5,0.75},
    {K_B1,1000,0.0,0.25},
    {K_B1, 200,1.0,0.75} }
};

//- - - - - - - - - - - - - -

void synth_instrument_drum::frequency(float freq)
{
  if (freq < 0.0) freq = 0;
  else if (freq > (AUDIO_SAMPLE_RATE/2)) freq = AUDIO_SAMPLE_RATE/2;

  wav_increment = (freq * (0x7fffffffLL/AUDIO_SAMPLE_RATE)) + 0.5;
}

void synth_instrument_drum::length(int32_t milliseconds)
{
  if(milliseconds < 0) return;
  if(milliseconds > 5000) milliseconds = 5000;

  int32_t len_samples = milliseconds*(AUDIO_SAMPLE_RATE/1000.0);

  env_decrement = (0x7fff0000/len_samples);
};

void synth_instrument_drum::secondMix(float level)
{
  // As level goes from 0.0 to 1.0,
  // second goes from 0 to 1/2 scale
  // first goes from full scale to half scale. 

  if (level < 0) level = 0;
  else if (level > 1.0) level = 1.0;

  noInterrupts();
    wav_amplitude2 = level * 0x3fff;
    wav_amplitude1 = 0x7fff - wav_amplitude2;
  interrupts();
}

void synth_instrument_drum::pitchMod(float depth)
{
  int32_t intdepth, calc;

  if (depth < 0) depth = 0;
  else if (depth > 1.0) depth = 1.0;

  // depth is float, 0.0..1.0
  // turn 0.0 to 1.0 into
  // 0x0 to 0x3fff
  intdepth = depth * 0x7fff;

  // turn it into 2.14, in range between -0.75 and 2.9999, with 0 at 0.5
  // It becomes the scalar for the modulation component of the phasor increment.
  if (intdepth < 0x4000)
  {
    // 0 to 0.5 becomes
    // -0x3000 (0xffffCfff) to 0 ()
    calc = ((0x4000 - intdepth) * 0x3000 ) >> 14;
    calc = -calc;
  }
  else
  {
    // 0.5 to 1.0 becomes
    // 0x00 to 0xbfa0
    calc = ((intdepth - 0x4000) * 0xc000) >> 14;
  }

  wav_pitch_mod = calc;
}

//- - - - - - - - - - - - - -

void synth_instrument_drum::startInstrument( ui_controls_struct *ui, int key ) 
{
  float f = drums[ui->effect_knob][key].frequency;
  switch (ui->band_knob)
  { case 0: f /= 2.0; break;
    case 1: f /= 1.5; break;
  //case 2: +0
    case 3: f *= 1.5; break;
    case 4: f *= 2.0; break;
  }
  frequency( f );
  length   ( drums[ui->effect_knob][key].length );
  pitchMod ( drums[ui->effect_knob][key].pitchMod );
//secondMix( drums[ui->effect_knob][key].secondMix );
  updateInstrument( ui ); // set secondMix from proximity
  
  noInterrupts();
    wav_phasor = 0;
    wav_phasor2 = 0;
    env_lin_current = 0x7fff0000;
  interrupts();
}

//- - - - - - - - - - - - - -

void synth_instrument_drum::stopInstrument()
{
  env_lin_current = 0;
}

//- - - - - - - - - - - - - -

void synth_instrument_drum::updateInstrument( ui_controls_struct *ui )
{
  // proximity effect
  if (ui->proximity > 5)
    secondMix( (float)ui->proximity / 255.0 );
  else 
    secondMix( 0.0 );
}

//- - - - - - - - - - - - - -

byte synth_instrument_drum::runInstrument( int32_t *buf )
{
  int32_t *end = buf + SOUND_BLOCK_SIZE;
  int32_t sin_l, sin_r, interp, mod, mod2, delta;
  int32_t interp2;
  int32_t index, scale;
  bool do_second;

  int32_t env_sqr_current; // the square of the linear value - inexpensive quasi exponential decay.

  if (env_lin_current < 0x0000ffff) return 0;

  // 50 is arbitrary threshold...
  // low values of second are inaudible, and we can save CPU cycles
  // by not calculating second when it's really quiet.
  do_second = (wav_amplitude2 > 50);

  while(buf < end)
  {
    // Do envelope first
    if (env_lin_current < 0x0000ffff)
    {
      // If envelope has expired, then stuff zeros into output buffer
      buf++;
    }
    else
    {
      env_lin_current -= env_decrement;
   // env_sqr_current = multiply_16tx16t(env_lin_current, env_lin_current) ;
      env_sqr_current = (env_lin_current >> 16) * (env_lin_current >> 16);

      // do wave second;
      wav_phasor  += wav_increment;

      // modulation changes how we use the increment
      // the increment will be scaled by the modulation amount.
      //
      // Pitch mod is in range [-0.75 .. 3.99999] in 2.14 format
      // Current envelope value gets scaled by mod depth.
      // Then phasor increment gets scaled by that.

   // mod = signed_multiply_32x16b((env_sqr_current), (wav_pitch_mod>>1)) >> 13;      
   // mod2 = signed_multiply_32x16b(wav_increment<<3, mod>>1);
      mod  = ((int64_t)env_sqr_current * ((int16_t)((wav_pitch_mod>>1) & 0xFFFF))) >> (13+16);      
      mod2 = (((int64_t)wav_increment<<3) * ((int16_t)((mod>>1) & 0xFFFF))) >> 16;
  
      wav_phasor += (mod2);
      wav_phasor &= 0x7fffffff;

      if (do_second)
      {
        // A perfect fifth uses increment of 1.5 times regular increment
        wav_phasor2 += wav_increment;
        wav_phasor2 += (wav_increment >> 1);
        wav_phasor2 += mod2;
        wav_phasor2 += (mod2 >> 1);
        wav_phasor2 &= 0x7fffffff;
      }
    
      // Phase to Sine lookup * interp:
      index = wav_phasor >> 23; // take top valid 8 bits
      sin_l = wavetable[0][index];
      sin_r = wavetable[0][index+1];

      // The fraction of the phasor in time we are between L and R
      // is the same as the fraction of the ampliture of that interval we should add 
      // to L.
      delta = sin_r-sin_l;
      scale = (wav_phasor >> 7) & 0xfFFF;
      delta = (delta * scale)>> 16;
      interp = sin_l + delta;
      
      if (do_second)
      {
        index = wav_phasor2 >> 23; // take top valid 8 bits
        sin_l = wavetable[0][index];
        sin_r = wavetable[0][index+1];

        delta = sin_r-sin_l;
        scale = (wav_phasor2 >> 7) & 0xFFFF;
        delta = (delta * scale)>> 16;
        interp2 = sin_l + delta;

        // Then scale and add the two waves
        interp2 = (interp2 * wav_amplitude2 ) >> 15;
        interp = (interp * wav_amplitude1) >> 15;
        interp = interp + interp2;
      }

   // *(p_wave++) = signed_multiply_32x16b(env_sqr_current, interp ) >> 15 ;
      *(buf++) += ((int64_t)env_sqr_current * (interp & 0xFFFF)) >> (15+16);
    }
  }
  return 1; // running
}

